import { of, combineLatest, from } from 'rxjs';
import { take, map, tap, filter, mergeMap, catchError, switchMap } from 'rxjs/operators';
import { moveItemInArray, transferArrayItem, DragDropModule } from '@angular/cdk/drag-drop';
import { NgModule } from '@angular/core';
import { AngularFirestoreModule } from '@angular/fire/firestore';

/** Helper method to get reference from path, the path can be either to a Document or Collection */
function getRefFromPath(path, ngFirestore) {
    const pathSegmentAmount = path.split('/').length;
    if (pathSegmentAmount % 2 === 0) { // even number means doc
        return ngFirestore.doc(path).ref;
    }
    else { // odd meaning collection
        return ngFirestore.collection(path).ref;
    }
}
/** If given the path to a Document it returns the parent AngularFirestoreCollection,
 * if given a path to a collection it just turns it into a AngularFirestoreCollection
 */
function getCollectionFsFromPath(path, ngFirestore) {
    const collectionPathSegments = path.split('/');
    const pathSegmentAmount = path.split('/').length;
    if (pathSegmentAmount % 2 === 0) { // even number means doc
        collectionPathSegments.pop(); // delete last
        return ngFirestore.collection(collectionPathSegments.join('/'));
    }
    else {
        return ngFirestore.collection(path);
    }
}
/**
 * Get AngularFirestoreDocument from path
 *
 * @param path the path to the AngularFirestoreDocument
 * @param ngFirestore the AngularFirestore object
 */
function getDocFsFromPath(path, ngFirestore) {
    return ngFirestore.doc(path);
}
/**
 * Returns the next higher level doc from either a collection or a doc path
 *
 * @param path the path to the AngularFirestoreDocument
 * @param ngFirestore the AngularFirestore object
 */
function getParentDocFsFromPath(path, ngFirestore) {
    let pathSegments = path.split('/');
    const pathSegmentAmount = pathSegments.length;
    if (pathSegmentAmount % 2 === 0) { // even number means doc, so we need to remove doc and the collection to get the parent doc
        pathSegments = pathSegments.slice(0, pathSegments.length - 2);
        return ngFirestore.doc(pathSegments.join('/'));
    }
    else { // odd, so its a colletion
        pathSegments = pathSegments.slice(0, pathSegments.length - 1);
        return ngFirestore.doc(pathSegments.join('/'));
    }
}
/**
 * Simple check if Document exists, returns true if Document exist
 *
 * @param docFs AngularFirestoreDocument to check if exist
 */
function isDocExist$(docFs) {
    return docFs.snapshotChanges().pipe(take(1), map((action) => {
        return action.payload.exists;
    }));
}
/**
 * Get the AngularFirestoreDocument object for a firestore DocumentReference
 *
 * AngularFirestore and Firestore uses different references for Documents and Collections.
 *
 * @param ref Firestore DocumentReference to Document
 * @param ngFirestore the AngularFirestore object
 */
function getDocFsFromRef(ref, ngFirestore) {
    return ngFirestore.doc(ref.path);
}
/**
 * Add createdDate to the object, if createdDate already exists then we do not overwrite it
 *
 * @param data data where the createdData will be added
 */
function addCreatedDate(data) {
    if ('createdDate' in data) {
        return data;
    }
    const createdDate = new Date();
    return Object.assign(Object.assign({}, data), { createdDate });
}
/**
 * Add modifiedDate to the object
 *
 * @param data data where the modifiedDate will be added
 */
function addModifiedDate(data) {
    const modifiedDate = new Date();
    return Object.assign(Object.assign({}, data), { modifiedDate });
}
/**
 * Add createdBy to the object
 *
 * @param createdBy profile, user or any type of data
 */
function addCreatedBy(data, createdBy) {
    return Object.assign(Object.assign({}, data), { createdBy });
}
/**
 * Firestore saves time as timestamps and javascript uses Date objects.
 * This functions helps convert the createdDate and modifiedDate from timestamp
 * to Date()
 *
 * @param data data that contains 'createdDate' and/or 'modifiedDate'
 */
function convertTimestampToDate(data) {
    if (data.hasOwnProperty('createdDate')) {
        data.createdDate = data.createdDate;
        data.createdDate = data.createdDate.toDate();
    }
    if (data.hasOwnProperty('modifiedDate')) {
        data.modifiedDate = data.modifiedDate;
        data.modifiedDate = data.modifiedDate.toDate();
    }
    return data;
}

/**
 * Action to be taken by listener if the document does not exist.
 */
var DocNotExistAction;
(function (DocNotExistAction) {
    /** returns a null object */
    DocNotExistAction[DocNotExistAction["RETURN_NULL"] = 0] = "RETURN_NULL";
    /** return all the extras such as ref, path and so on but no data, kinda just ignores that the doc isn't there */
    DocNotExistAction[DocNotExistAction["RETURN_ALL_BUT_DATA"] = 1] = "RETURN_ALL_BUT_DATA";
    /** do not return at all until it does exist */
    DocNotExistAction[DocNotExistAction["FILTER"] = 2] = "FILTER";
    /** return doc not found error 'doc_not_found' */
    DocNotExistAction[DocNotExistAction["THROW_DOC_NOT_FOUND"] = 3] = "THROW_DOC_NOT_FOUND";
})(DocNotExistAction || (DocNotExistAction = {}));
/**
 * Main Class.
 *
 *
 *
 */
class AngularFirestoreDeep {
    /**
     * Constructor for AngularFirestoreWrapper
     *
     * @param ngFirestore The AngularFirestore injected from an Angular Class
     * @param defaultDocId The default name of any Document when no name is given
     */
    constructor(ngFirestore, defaultDocId = 'data') {
        this.ngFirestore = ngFirestore;
        this.defaultDocId = defaultDocId;
    }
    /* ----------  LISTEN -------------- */
    /**
     * Same as AngularFirestoreDocument.snapshotChanges but it adds the properties in FirebaseDbItem
     * and also allows for to choose action to take when document does not exist
     *
     * Important to understand this is will trigger for every change/update on the document we are listening to.
     *
     * @param docFs AngularFirestoreDocument that will be listend to
     * @param actionIfNotExist Action to take if document does not exist
     */
    listenForDoc$(docFs, actionIfNotExist = DocNotExistAction.RETURN_ALL_BUT_DATA) {
        /**
         * Returns an observable that will emit whenever the ref changes in any way.
         * Also adds the id and ref to the object.
         */
        const res$ = docFs.snapshotChanges().pipe(
        // mergeMap((action) => {
        //   if (!action.payload.exists) { return throwError('Document Does not exists'); }
        //   else { return of(action); }
        // }),
        tap((action) => {
            if (!action.payload.exists && actionIfNotExist === DocNotExistAction.THROW_DOC_NOT_FOUND) {
                throw new Error('doc_not_found');
            }
        }), filter((action) => {
            return !(!action.payload.exists && actionIfNotExist === DocNotExistAction.FILTER);
        }), map((action) => {
            if (action.payload.exists || actionIfNotExist === DocNotExistAction.RETURN_ALL_BUT_DATA) {
                const data = action.payload.data();
                // const payload: DocumentSnapshotExists<DocumentData> = action.payload as DocumentSnapshotExists<DocumentData>;
                // @ts-ignore
                const id = action.payload.id;
                // @ts-ignore
                const docRef = action.payload.ref;
                const path = docRef.path;
                const _docFs = this.ngFirestore.doc(path);
                const isExists = action.payload.exists;
                // const afDocRef = this.fireStore.doc<A>(docRef.path) as AngularFirestoreDocument;  // TODO make sure this also copies the query
                return Object.assign(Object.assign({}, data), { id, path, ref: docRef, docFs: _docFs, isExists });
            }
            else if (actionIfNotExist === DocNotExistAction.RETURN_NULL) { /* doc doesn't exist */
                return null;
            }
        }), map((data) => {
            if (data != null) {
                return convertTimestampToDate(data);
            }
            else {
                return data;
            }
        }));
        return this.handleRes$(res$, 'listenForDoc$', { docFs });
    }
    /**
     * Same as AngularFirestoreCollection.snapshotChanges but it adds the properties in FirebaseDbItem.
     *
     * Important to understand this is will trigger for every change/update on any of the documents we are listening to.
     * That means that if any document we are listening to is changed the entire object will be triggered containing the updated data.
     *
     *
     *    Example usage.
     *
     *    ngFirestoreDeep: AngularFirestoreDeep;  //  AngularFirestoreDeep variable
     *    restaurantCollectionFs = this.ngFireStore.collection('restaurants'); // AngularFirestoreCollectionRef to restaurants
     *
     *    constructor(private ngFireStore: AngularFirestore) {
     *        this.ngFirestoreDeep = new AngularFirestoreDeep(ngFireStore);  //  initialize AngularFireStoreDeep with AngularFirestore
     *    }
     *
     *    listenForRestaurants$(): Observable<RestaurantItem[]> {
     *        return this.ngFirestoreDeep.listenForCollection$<RestaurantItem>(this.restaurantCollectionFs);
     *    }
     *
     *    If you do not wish to listen for changes and only care about getting the values once
     *
     *    getRestaurants$(): Observable<RestaurantItem[]> {
     *        return this.ngFirestoreDeep.listenForCollection$<RestaurantItem>(this.restaurantCollectionFs).pipe(
     *          take(1)
     *        );
     *    }
     *
     *
     * @param collectionFs the AngularFirestoreCollection which will be listened to
     * @param documentChangeTypes list of DocumentChangeType that will be listened to, if null listen to all
     */
    listenForCollection$(collectionFs, documentChangeTypes) {
        /**
         * Returns an observable that will emit whenever the ref changes in any way.
         * Also adds the id and ref to the object.
         */
        if (documentChangeTypes === undefined || documentChangeTypes === null || documentChangeTypes.length <= 0) {
            documentChangeTypes = ['added', 'removed', 'modified'];
        }
        const res$ = collectionFs.snapshotChanges().pipe(map((actions) => actions.filter(a => {
            return documentChangeTypes.includes(a.type);
        })), map((actions) => actions.map(a => {
            const data = a.payload.doc.data();
            const id = a.payload.doc.id;
            const docRef = a.payload.doc.ref;
            const path = docRef.path;
            const _docFs = this.ngFirestore.doc(path);
            // const afDocRef = this.fireStore.doc<A>(docRef.path) as AngularFirestoreDocument; // TODO make sure this also copies the query
            return Object.assign(Object.assign({}, data), { id, path, ref: docRef, docFs: _docFs });
        })), map((datas) => datas.map(data => {
            return convertTimestampToDate(data);
        })));
        return this.handleRes$(res$, 'listenForCollection$', { collectionFs, listenToTypes: documentChangeTypes });
    }
    /**
     *
     * Allows for listening to documents and collections n deep up to the firestore max of 100 levels.
     *
     * Triggers for any change in any document that is listened to.
     *
     *
     * E.x:
     *      const subCollectionQueries: SubCollectionQuery[] = [
     *         { name: 'data' },
     *         { name: 'secure' },
     *         { name: 'variants' },
     *         { name: 'images',
     *           queryFn: ref => ref.orderBy('index'),
     *           collectionWithNames: [
     *             { name: 'secure'}
     *           ]
     *         },
     *     ];
     *
     *     this.listenForDocAndSubCollections<Product>(docFs, collections)
     *
     * Wrapper for listenForDocDeepRecursiveHelper$ so that we can cast the return to the correct type
     * All logic is in listenForDocDeepRecursiveHelper$.
     *
     * @param docFs - a docFs with potential queryFn
     * @param subCollectionQueries - see example
     */
    listenForDocDeep$(docFs, subCollectionQueries = []) {
        const res$ = this.listenForDocDeepRecursiveHelper$(docFs, subCollectionQueries).pipe(map(data => data));
        return this.handleRes$(res$, 'listenForDocDeep$', { docFs, subCollectionQueries }, null, 'Failed To read doc');
    }
    /**
     * DO NOT CALL THIS METHOD, meant to be used solely by listenForDocAndSubCollections$
     */
    listenForDocDeepRecursiveHelper$(docFs, subCollectionQueries = [], actionIfNotExist = DocNotExistAction.RETURN_NULL) {
        /* Listen for the docFs*/
        return this.listenForDoc$(docFs, actionIfNotExist).pipe(
        //
        // catchError(err => {
        //   if (err === 'Document Does not exists') { return of(null); } /* if a doc is deleted we just skip it */
        //   else { throw err; }
        // }),
        filter(doc => doc !== null), mergeMap((item) => {
            if (item === null) {
                return of(item);
            }
            if (subCollectionQueries.length <= 0) {
                return of(item);
            }
            const collectionListeners = [];
            /* Iterate over each sub collection we have given and create collection listeners*/
            subCollectionQueries.forEach(subCollectionQuery => {
                // console.log(subCollectionQuery);
                const collectionFs = docFs.collection(subCollectionQuery.name, subCollectionQuery.queryFn);
                // const collectionFs = this.firestore.docFs(item.path).subCollectionQuery(subCollectionQuery.name, subCollectionQuery.queryFn);
                const collectionListener = this.listenForCollection$(collectionFs, null).pipe(
                // tap(d => console.log(d)),
                // filter(docs => docs.length > 0), // skip empty collections or if the subCollectionQuery doesnt exist
                /* Uncomment to see data on each update */
                // tap(d => console.log(d)),
                /* Listen For and Add any Potential Sub Docs*/
                mergeMap((docs) => {
                    if (!subCollectionQuery.subCollectionQueries) {
                        return of(docs);
                    }
                    const docListeners = [];
                    docs.forEach(d => {
                        const subDocFs = this.ngFirestore.doc(d.path);
                        const subDocAndCollections$ = this.listenForDocDeepRecursiveHelper$(subDocFs, subCollectionQuery.subCollectionQueries).pipe(
                        // tap(subDoc => console.log(subDoc)),
                        // filter(subDoc => subDoc !== null),
                        map((subDoc) => {
                            return Object.assign(Object.assign({}, d), subDoc);
                        }));
                        docListeners.push(subDocAndCollections$);
                    });
                    // console.log(docListeners);
                    if (docListeners.length <= 0) {
                        return of([]);
                    } /* subCollectionQuery is empty or doesnt exist */
                    return combineLatest(docListeners).pipe(
                    // tap(val => console.log(val))
                    );
                }), /* End of Listening for sub docs */ 
                // tap(val => console.log(val)),
                /* If docs.length === 1 and the id is defaultDocId or the given docId it means we are in a sub subCollectionQuery
                and we only care about the data. So we remove the array and just make it one object with the
                subCollectionQuery name as key and docs[0] as value */
                map((docs) => {
                    const docId = subCollectionQuery.docId !== undefined ? subCollectionQuery.docId : this.defaultDocId;
                    if (docs.length === 1 && docs[0].id === docId) {
                        return { [subCollectionQuery.name]: docs[0] };
                    }
                    else {
                        return { [subCollectionQuery.name]: docs };
                    }
                }));
                collectionListeners.push(collectionListener);
            });
            /* Finally return the combined collection listeners*/
            return combineLatest(collectionListeners).pipe(map((datas) => {
                const datasMap = {};
                datas.forEach((collection) => {
                    for (const [key, value] of Object.entries(collection)) {
                        datasMap[key] = value;
                    }
                });
                return datasMap;
            }), map((data) => {
                return Object.assign(Object.assign({}, item), data);
            }));
        }));
    }
    /**
     * Listens for collections inside collections with the same name to an unlimited depth and returns all of it as an array.
     */
    listenForCollectionRecursively$(path, collectionKey, orderKey) {
        let ref;
        if (orderKey != null) {
            ref = this.ngFirestore.collection(path, r => r.orderBy(orderKey));
        }
        else {
            ref = this.ngFirestore.collection(path);
        }
        return this.listenForCollection$(ref, null).pipe(mergeMap((items) => {
            if (items.length <= 0) {
                return of([]);
            } // TODO  perhaps make this throw an error so that we can skip it
            // if (items.length <= 0) { throwError('No more '); }
            const nextLevelObs = [];
            for (const item of items) {
                const nextLevelPath = this.ngFirestore.doc(item.path).collection(collectionKey).ref.path; // one level deeper
                const nextLevelItems$ = this.listenForCollectionRecursively$(nextLevelPath, collectionKey, orderKey).pipe(map((nextLevelItems) => {
                    if (nextLevelItems.length > 0) {
                        return Object.assign(Object.assign({}, item), { [collectionKey]: nextLevelItems });
                    }
                    else {
                        return Object.assign({}, item);
                    } // dont include an empty array
                }));
                nextLevelObs.push(nextLevelItems$);
            }
            return combineLatest(nextLevelObs).pipe(tap(val => console.log(val)));
        }));
    }
    /* ----------  ADD -------------- */
    /**
     * A replacement/extension to the AngularFirestoreCollection.add.
     * Does the same as AngularFirestoreCollection.add but can also add createdDate and modifiedDate and returns
     * the data with the added properties in FirebaseDbItem
     *
     * Used internally
     *
     * @param data the data to be added to the document, cannot contain types firestore won't allow
     * @param collectionFs the AngularFirestoreCollection where the document should be added
     * @param isAddDates if true adds modifiedDate and createdDate to the data
     * @param id if given the added document will be given this id, otherwise a random unique id will be used.
     */
    addSimple$(data, collectionFs, isAddDates = true, id) {
        return of(null).pipe(mergeMap(() => {
            let res$;
            if (isAddDates) {
                data = addCreatedDate(data);
                data = addModifiedDate(data);
            }
            if (id !== undefined) {
                res$ = from(collectionFs.doc(id).set(data));
            }
            else {
                res$ = from(collectionFs.add(data));
            }
            res$ = res$.pipe(
            // tap(() => this.snackBar.open('Success', 'Added', {duration: 1000})),
            // tap(ref => console.log(ref)),
            // tap(() => console.log(data)),
            map((ref) => {
                if (id === undefined) {
                    return Object.assign(Object.assign({}, data), { id: ref.id, path: ref.path, ref, docFs: this.ngFirestore.doc(ref.path) });
                }
                else {
                    const path = collectionFs.ref.path + '/' + id;
                    return Object.assign(Object.assign({}, data), { id, path, docFs: this.ngFirestore.doc(path) });
                }
            }));
            return this.handleRes$(res$, 'add$', { data, collectionFs, id }, 'Document Added Successfully', 'Document failed to add').pipe(take(1) /* No need to unsub */);
        }));
    }
    /**
     * Add document to firestore and split it up into sub collection.
     *
     * @param data the data to be saved
     * @param collectionFs AngularFirestoreCollection reference to where on firestore the item should be saved
     * @param subCollectionWriters see documentation for SubCollectionWriter for more details on how these are used
     * @param isAddDates if true 'createdDate' and 'modifiedDate' is added to the data
     * @param docId If a docId is given it will use that specific id when saving the doc, if no docId is given a ranom id will be used.
     */
    addDeep$(data, collectionFs, subCollectionWriters = [], isAddDates = true, docId) {
        const split = this.splitDataIntoCurrentDocAndSubCollections(data, subCollectionWriters);
        const currentDoc = split.currentDoc;
        const subCollections = split.subCollections;
        // console.log(currentDoc, subCollections);
        // console.log(Object.entries(subCollections).keys(), Object.entries(subCollections).values());
        const res$ = this.addSimple$(currentDoc, collectionFs, isAddDates, docId).pipe(
        // tap(val => console.log(val)),
        /* Add Sub/sub collections*/
        mergeMap((currentData) => {
            const subWriters = [];
            for (const [subCollectionKey, subCollectionValue] of Object.entries(subCollections)) {
                let subSubCollectionWriters; // undefined in no subCollectionWriters
                let subDocId;
                if (subCollectionWriters) {
                    subSubCollectionWriters = subCollectionWriters.find(subColl => subColl.name === subCollectionKey).subCollectionWriters;
                    subDocId = subCollectionWriters.find(subColl => subColl.name === subCollectionKey).docId;
                }
                const subCollectionFs = this.ngFirestore.doc(currentData.path).collection(subCollectionKey);
                /* Handle array and object differently
                * For example if array and no docId is given it means we should save each entry as a separate doc.
                * If a docId is given should save it using that docId under a single doc.
                * If not an array it will always be saved as a single doc, using this.defaultDocId as the default docId if none is given */
                if (Array.isArray(subCollectionValue)) {
                    if (subDocId !== undefined) { /* not undefined so save it as a single doc under that docId */
                        /* the pipe only matters for the return subCollectionValue not for writing the data */
                        const subWriter = this.addDeep$(subCollectionValue, subCollectionFs, subSubCollectionWriters, isAddDates, subDocId).pipe(map(item => {
                            // return {[key]: item};
                            return { key: subCollectionKey, value: item }; /* key and subCollectionValue as separate k,v properties */
                        }));
                        subWriters.push(subWriter);
                    }
                    else { /* docId is undefined so we save each object in the array separate */
                        subCollectionValue.forEach((arrayValue) => {
                            /* the pipe only matters for the return subCollectionValue not for writing the data */
                            const subWriter = this.addDeep$(arrayValue, subCollectionFs, subSubCollectionWriters, isAddDates).pipe(map(item => {
                                // return {[key]: [item]};
                                /* key and subCollectionValue as separate k,v properties -- subCollectionValue in an array */
                                return { key: subCollectionKey, value: [item] };
                            }));
                            subWriters.push(subWriter);
                        });
                    }
                }
                else { /* Not an array so a single Object*/
                    subDocId = subDocId !== undefined ? subDocId : this.defaultDocId;
                    /* the pipe only matters for the return subCollectionValue not for writing the data */
                    const subWriter = this.addDeep$(subCollectionValue, subCollectionFs, subSubCollectionWriters, isAddDates, subDocId).pipe(map(item => {
                        // return {[key]: item};
                        return { key: subCollectionKey, value: item }; /* key and subCollectionValue as separate k,v properties */
                    }));
                    subWriters.push(subWriter);
                }
            } /* end of iteration */
            if (subWriters.length > 0) { /* if subWriters.length > 0 it means we need to handle the subwriters */
                /* the pipe only matters for the return value not for writing the data */
                return combineLatest(subWriters).pipe(
                // tap(sub => console.log(sub)),
                // TODO super duper ugly way of joining the data together but I cannot think of a better way..also it doesnt really matter.
                // TODO The ugliness only relates to how the return object looks after we add, it has no effect on how the object is saved on
                // TODO firestore.
                map((docDatas) => {
                    const groupedData = {};
                    docDatas.forEach((doc) => {
                        const key = doc.key;
                        const value = doc.value;
                        /* if groupedData has the key already it means that the several docs have the same key..so an array */
                        if (groupedData.hasOwnProperty(key) && Array.isArray(groupedData[key])) {
                            /* groupedData[key] must be an array since it already exist..add this doc.value to the array */
                            groupedData[key].push(value[0]);
                        }
                        else {
                            groupedData[key] = value;
                        }
                    });
                    return groupedData;
                }), 
                // tap(groupedData => console.log(groupedData)),
                map((groupedData) => {
                    return Object.assign(Object.assign({}, currentData), groupedData);
                }));
            }
            else {
                return of(currentData);
            }
        }));
        return this.handleRes$(res$, 'addDeep$', { data, collectionFs, subCollectionWriters, docId }, 'Document Added Successfully', 'Document failed to add').pipe(
        // tap(d => console.log(d)),
        take(1) /* just to make sure it only emits once */);
    }
    /* ----------  EDIT -------------- */
    /**
     * clean FirestoreBaseItem properties from the data.
     * Usually done if you wish to save the data to firestore, since some FirestoreBaseItem properties are of non allowed types.
     *
     * Goes through each level and removes DbItemExtras
     * In case you wish to save the data
     *
     * @param data data to be cleaned
     * @param subCollectionWriters if the document has child documents the SubCollectionWriters are needed to locate them
     */
    cleanExtrasFromData(data, subCollectionWriters = []) {
        // const dataToBeCleaned = cloneDeep(data); /* clone data so we dont modify the original */
        // const dataToBeCleaned = data;
        return this.removeDataExtrasRecursiveHelper(data, subCollectionWriters);
    }
    /**
     * Recursive method to clean FirestoreBaseItem properties from the dbItem
     *
     * @param dbItem the data to be cleaned
     * @param subCollectionWriters list of SubCollectionWriters to handle sub collections
     */
    removeDataExtrasRecursiveHelper(dbItem, subCollectionWriters = []) {
        // const extraPropertyNames: string[] = Object.getOwnPropertyNames(new DbItemExtras());
        const extraPropertyNames = ['id', 'path', 'ref', 'docFs', 'isExists'];
        /* Current level delete */
        for (const extraPropertyName of extraPropertyNames) {
            delete dbItem[extraPropertyName];
        }
        subCollectionWriters.forEach(col => {
            if (Array.isArray(dbItem[col.name])) { /* property is array so will contain multiple docs */
                const docs = dbItem[col.name];
                docs.forEach((d, i) => {
                    if (col.subCollectionWriters) {
                        this.removeDataExtrasRecursiveHelper(d, col.subCollectionWriters);
                    }
                    else {
                        /*  */
                        for (const extraPropertyName of extraPropertyNames) {
                            delete dbItem[col.name][i][extraPropertyName];
                        }
                    }
                });
            }
            else { /* not an array so a single doc*/
                if (col.subCollectionWriters) {
                    this.removeDataExtrasRecursiveHelper(dbItem[col.name], col.subCollectionWriters);
                }
                else {
                    for (const extraPropertyName of extraPropertyNames) {
                        delete dbItem[col.name][extraPropertyName];
                    }
                }
            }
        });
        return dbItem;
    }
    /**
     * Firestore doesn't allow you do change the name or move a doc directly so you will have to create a new doc under the new name
     * and then delete the old doc.
     * returns the new doc once the delete is done.
     *
     * @param docFs AngularFirestoreDocument to have its id changed
     * @param newId the new id
     * @param subCollectionQueries if the document has child documents the subCollectionQueries are needed to locate them
     * @param subCollectionWriters if the document has child documents the SubCollectionWriters are needed to add them back
     */
    changeDocId$(docFs, newId, subCollectionQueries = [], subCollectionWriters = []) {
        const collectionFs = this.ngFirestore.collection(docFs.ref.parent);
        const res$ = this.listenForDocDeep$(docFs, subCollectionQueries).pipe(take(1), map((oldData) => this.cleanExtrasFromData(oldData, subCollectionWriters)), mergeMap((oldData) => {
            return this.addDeep$(oldData, collectionFs, subCollectionWriters, false, newId).pipe(/* add the data under id*/ mergeMap(newData => {
                return this.deleteDeep$(docFs, subCollectionQueries).pipe(map(() => newData) /* keep the new data */);
            }));
        }), catchError(err => {
            console.log('Failed to Change Doc Id: ' + err);
            throw err;
        }));
        return this.handleRes$(res$, 'changeDocName$', { docFs, subCollectionQueries, subCollectionWriters, newName: newId }, 'Failed to Change Doc Name');
    }
    /**
     * Update document and child documents
     *
     * Be careful when updating a document of any kind since we allow partial data there cannot be any type checking prior to update
     * so its possible to introduce spelling mistakes on attributes and so forth
     *
     * @param data the data that is to be added or updated { [field: string]: any }
     * @param docFs AngularFirestoreDocument to be updated
     * @param subCollectionWriters if the data contains properties that should be placed in child collections and documents specify that here
     * @param isAddModifiedDate if true the modifiedDate property is added/updated on the affected documents
     */
    updateDeep$(data, docFs, subCollectionWriters = [], isAddModifiedDate = true) {
        if (subCollectionWriters == null || subCollectionWriters.length === 0) {
            return this.updateSimple$(data, docFs, isAddModifiedDate); // no subCollectionWriters so just do a simple update
        }
        const batch = this.updateDeepToBatchHelper(data, docFs, subCollectionWriters, isAddModifiedDate);
        const res$ = this.batchCommit$(batch);
        return this.handleRes$(res$, 'updateDeep$', { data, docFs, subCollectionWriters }, 'Updated', 'Failed to Update Document').pipe(take(1) /* no need to unsub */);
    }
    /**
     * DO NOT CALL THIS METHOD, used by update deep
     */
    updateDeepToBatchHelper(data, docFs, subCollectionWriters = [], isAddModifiedDate = true, batch) {
        if (batch === undefined) {
            batch = this.ngFirestore.firestore.batch();
        }
        if (isAddModifiedDate) {
            data = addModifiedDate(data);
        }
        const split = this.splitDataIntoCurrentDocAndSubCollections(data, subCollectionWriters);
        const currentDoc = split.currentDoc;
        const subCollections = split.subCollections;
        // console.log(currentDoc, subCollections);
        batch.update(docFs.ref, currentDoc);
        for (const [subCollectionKey, subDocUpdateValue] of Object.entries(subCollections)) {
            let subSubCollectionWriters; // undefined in no subCollectionWriters
            let subDocId;
            if (subCollectionWriters) {
                subSubCollectionWriters = subCollectionWriters.find(subColl => subColl.name === subCollectionKey).subCollectionWriters;
                subDocId = subCollectionWriters.find(subColl => subColl.name === subCollectionKey).docId;
            }
            subDocId = subDocId !== undefined ? subDocId : this.defaultDocId; /* Set default if none given */
            const subDocFs = docFs.collection(subCollectionKey).doc(subDocId);
            batch = this.updateDeepToBatchHelper(subDocUpdateValue, subDocFs, subSubCollectionWriters, isAddModifiedDate, batch);
        }
        return batch;
    }
    /**
     * Used internally for updates that doesn't affect child documents
     */
    updateSimple$(data, docFs, isAddModifiedDate = true) {
        /**
         * transforms the promise to an observable
         */
        return of(null).pipe(mergeMap(() => {
            let res$;
            if (isAddModifiedDate) {
                data = addModifiedDate(data);
            }
            res$ = from(docFs.update(data));
            return this.handleRes$(res$, 'update$', { data, docFs }, 'Updated', 'Failed to Update Document').pipe(
            // take(1) /* no need to unsub */
            );
        }), take(1) /* just to make sure it only emits once */);
    }
    /**
     * Update the firestore document by path
     * Convenience method in case we do not have direct access to the AngularFirestoreDocument reference
     *
     * @param data the data that is to be added or updated { [field: string]: any }
     * @param path A string representing the path of the referenced document (relative to the root of the database).
     * @param subCollectionWriters if the data contains properties that should be placed in child collections and documents specify that here
     * @param isAddModifiedDate if true the modifiedDate property is added/updated on the affected documents
     */
    updateDeepByPath$(data, path, subCollectionWriters = [], isAddModifiedDate = true) {
        const doc = this.ngFirestore.doc(path);
        return this.updateDeep$(data, doc, subCollectionWriters, isAddModifiedDate);
    }
    /**
     * Update/ add data to the firestore documents
     *
     * @param docsFs list of AngularFirestoreDocument to be have their data updated
     * @param data data to add/update
     * @param isAddModifiedDate if true the modifiedDate is added/updated
     */
    updateMultiple$(docsFs, data, isAddModifiedDate = true) {
        const batch = this.ngFirestore.firestore.batch();
        if (isAddModifiedDate) {
            data = addModifiedDate(data);
        }
        docsFs.forEach((doc) => {
            batch.update(doc.ref, data);
        });
        return this.batchCommit$(batch);
    }
    /**
     * Update the firestore documents by paths
     *
     * @param path list of paths of documents to be have their data updated
     * @param data data to add/update
     * @param isAddModifiedDate if true the modifiedDate is added/updated
     */
    updateMultipleByPaths$(paths, data, isAddModifiedDate = true) {
        const docsFs = [];
        paths.forEach(path => {
            docsFs.push(this.ngFirestore.doc(path));
        });
        return this.updateMultiple$(docsFs, data, isAddModifiedDate);
    }
    /* Move Item in Array */
    moveItemInArray$(array, fromIndex, toIndex) {
        // console.log(fromIndex, toIndex);
        if (fromIndex == null || toIndex == null || fromIndex === toIndex || array.length <= 0) { // we didnt really move anything
            return of(null);
        }
        const batch = this.getBatchFromMoveItemInIndexedDocs(array, fromIndex, toIndex);
        return this.batchCommit$(batch);
    }
    getBatchFromMoveItemInIndexedDocs(array, fromIndex, toIndex, useCopy = false) {
        /**
         * Moved item within the same list so we need to update the index of all items in the list;
         * Use a copy if you dont wish to update the given array, for example when you watch to just listen for the change of the db..
         * The reason to not do this is because it takes some time for the db to update and it looks better if the list updates immediately.
         */
        const lowestIndex = Math.min(fromIndex, toIndex);
        const batch = this.ngFirestore.firestore.batch();
        if (fromIndex == null || toIndex == null || fromIndex === toIndex) { // we didnt really move anything
            return batch;
        }
        let usedArray;
        if (useCopy) {
            usedArray = Object.assign([], array);
        }
        else {
            usedArray = array;
        }
        // console.log(usedArray, array);
        moveItemInArray(usedArray, fromIndex, toIndex);
        const listSliceToUpdate = usedArray.slice(lowestIndex);
        let i = lowestIndex;
        for (const item of listSliceToUpdate) {
            const ref = getRefFromPath(item.path, this.ngFirestore);
            batch.update(ref, { index: i });
            i++;
        }
        return batch;
    }
    getBatchFromTransferItemInIndexedDocs(previousArray, currentArray, previousIndex, currentIndex, additionalDataUpdateOnMovedItem, useCopy = false) {
        /**
         * Used mainly for drag and drop scenarios where we drag an item from one array to another and the the docs have an index attribute.
         */
        const batch = this.ngFirestore.firestore.batch();
        let usedPreviousArray;
        let usedCurrentArray;
        if (useCopy) {
            usedPreviousArray = Object.assign([], previousArray);
            usedCurrentArray = Object.assign([], currentArray);
        }
        else {
            usedPreviousArray = previousArray;
            usedCurrentArray = currentArray;
        }
        transferArrayItem(usedPreviousArray, usedCurrentArray, previousIndex, currentIndex);
        if (additionalDataUpdateOnMovedItem !== undefined) {
            const movedItem = currentArray[currentIndex];
            const movedPartRef = getRefFromPath(movedItem.path, this.ngFirestore);
            batch.update(movedPartRef, additionalDataUpdateOnMovedItem);
        }
        const currentArraySliceToUpdate = usedCurrentArray.slice(currentIndex);
        let i = currentIndex;
        for (const item of currentArraySliceToUpdate) {
            const ref = getRefFromPath(item.path, this.ngFirestore);
            batch.update(ref, { index: i });
            i++;
        }
        const prevArraySliceToUpdate = usedPreviousArray.slice(previousIndex);
        i = previousIndex;
        for (const item of prevArraySliceToUpdate) {
            const ref = getRefFromPath(item.path, this.ngFirestore);
            batch.update(ref, { index: i });
            i++;
        }
        return batch;
    }
    getBatchFromDeleteItemInIndexedDocs(array) {
        /**
         * Run this on collections with a fixed order using an index: number attribute;
         * This will update that index with the index in the collectionData, so it should be sorted by index first.
         * Basically needs to be run after a delete
         */
        const batch = this.ngFirestore.firestore.batch();
        array.forEach((item, index) => {
            if (item.index !== index) {
                const ref = getRefFromPath(item.path, this.ngFirestore);
                batch.update(ref, { index });
            }
        });
        return batch;
    }
    updateIndexAfterDeleteInIndexedDocs(array) {
        const batch = this.getBatchFromDeleteItemInIndexedDocs(array);
        return this.batchCommit$(batch);
    }
    /* ----------  DELETE -------------- */
    /**
     * Delete document
     * Used internally
     *
     * @param docFs AngularFirestoreDocument that is to be deleted
     */
    deleteSimple$(docFs) {
        return of(null).pipe(mergeMap(() => {
            let res$;
            res$ = from(docFs.delete());
            return this.handleRes$(res$, 'deleteSimple$', { docFs }, 'Deleted', 'Failed to Delete Document').pipe(take(1) /* no need to unsub*/);
        }), take(1));
    }
    /**
     * Delete Document and child documents
     *
     * @param docFs AngularFirestoreDocument that is to be deleted
     * @param subCollectionQueries if the document has child documents the subCollectionQueries are needed to locate them
     */
    deleteDeep$(docFs, subCollectionQueries = []) {
        if (subCollectionQueries == null || subCollectionQueries.length === 0) {
            return this.deleteSimple$(docFs);
        }
        // console.log('deleteDeep');
        const res$ = this.getFirestoreDocumentsDeep$(docFs, subCollectionQueries).pipe(switchMap((docList) => this.deleteMultiple$(docList)), catchError((err) => {
            if (err === 'Document Does not exists') {
                return of(null);
            }
            else {
                throw err;
            }
        }), take(1));
        return this.handleRes$(res$, 'deleteDeep$', { docFs, subCollectionQueries }, 'Deleted', 'Failed to Delete');
    }
    /**
     * Delete Documents
     *
     * @param docsFs - A list of AngularFireDocument that are to be deleted
     */
    deleteMultiple$(docsFs) {
        const batch = this.ngFirestore.firestore.batch();
        docsFs.forEach((doc) => {
            batch.delete(doc.ref);
        });
        const res$ = this.batchCommit$(batch);
        return this.handleRes$(res$, 'deleteMultiple$', docsFs.map(doc => doc.ref.path), 'Deleted', 'Failed to Delete Multiple');
    }
    deleteMultipleByPaths$(paths) {
        const docsFs = [];
        paths.forEach(path => {
            docsFs.push(this.ngFirestore.doc(path));
        });
        return this.deleteMultiple$(docsFs);
    }
    /**
     * Delete Documents and child documents
     *
     * @param docsFs - A list of AngularFireDocument that are to be deleted
     * @param subCollectionQueries if the document has child documents the subCollectionQueries are needed to locate them
     */
    deleteMultipleDeep$(docsFs, subCollectionQueries = []) {
        if (subCollectionQueries == null || subCollectionQueries.length === 0) {
            return this.deleteMultiple$(docsFs);
        }
        const mainDocLists$ = [];
        docsFs.forEach(doc => {
            const docList$ = this.getFirestoreDocumentsDeep$(doc, subCollectionQueries);
            mainDocLists$.push(docList$);
        });
        return combineLatest(mainDocLists$).pipe(
        // tap(lists => console.log(lists)),
        map((lists) => {
            let mainDocFsList = [];
            lists.forEach(list => {
                mainDocFsList = mainDocFsList.concat(list);
            });
            return mainDocFsList;
        }), 
        // tap(lists => console.log(lists)),
        switchMap((docFsList) => this.deleteMultiple$(docFsList)), catchError((err) => {
            if (err === 'Document Does not exists') {
                return of(null);
            }
            else {
                throw err;
            }
        }));
    }
    /**
     * Delete firestore document by path
     * Convenience method in case we do not have direct access to the AngularFirestoreDocument reference
     *
     * @param path A string representing the path of the referenced document (relative to the root of the database).
     * @param subCollectionQueries if the document has child documents the subCollectionQueries are needed to locate them
     */
    deleteDocByPath$(path, subCollectionQueries = []) {
        const doc = this.ngFirestore.doc(path);
        return this.deleteDeep$(doc, subCollectionQueries);
    }
    /**
     * Delete document by FirestoreItem
     *
     * Convenience method that works the same as deleteDeep$ but takes a FirestoreItem to be deleted
     *
     * @param item FirestoreItem to be deleted
     * @param subCollectionQueries if the document has child documents the subCollectionQueries are needed to locate them
     */
    deleteDeepByItem$(item, subCollectionQueries = []) {
        // console.log('deleteDeepByItem$');
        const docsFs = this.getAngularFirestoreDocumentsFromDbItem(item, subCollectionQueries);
        const res$ = this.deleteMultiple$(docsFs).pipe(catchError((err) => {
            if (err === 'Document Does not exists') {
                return of(null);
            }
            else {
                throw err;
            }
        }), take(1));
        return this.handleRes$(res$, 'deleteDeepByItem$', { item, subCollectionQueries }, 'Deleted', 'Failed to Delete');
    }
    /* ----------  OTHER -------------- */
    /**
     * Returns an Observable containing a list of AngularFirestoreDocument found under the given docFs using the SubCollectionQuery[]
     * Mainly used to delete a docFs and its sub docs
     * @param docFs: AngularFirestoreDocument
     * @param subCollectionQueries: SubCollectionQuery[]
     */
    getFirestoreDocumentsDeep$(docFs, subCollectionQueries = []) {
        return this.listenForDocDeep$(docFs, subCollectionQueries).pipe(take(1), map(item => this.getPathsFromDbItemDeepRecursiveHelper(item, subCollectionQueries)), 
        // tap(pathList => console.log(pathList)),
        map((pathList) => {
            const docFsList = [];
            pathList.forEach(path => docFsList.push(this.ngFirestore.doc(path)));
            return docFsList;
        }));
    }
    /**
     * Used by deleteDeepByItem$ to get all the AngularFirestoreDocuments to be deleted
     * including child documents using SubCollectionQueries
     *
     * Internal use
     * @param dbItem FirestoreItem from where we get the AngularFirestoreDocuments
     * @param subCollectionQueries if the dbItem has child documents the subCollectionQueries are needed to locate them
     */
    getAngularFirestoreDocumentsFromDbItem(dbItem, subCollectionQueries = []) {
        const pathList = this.getPathsFromDbItemDeepRecursiveHelper(dbItem, subCollectionQueries);
        const docFsList = [];
        pathList.forEach(path => docFsList.push(this.ngFirestore.doc(path)));
        return docFsList;
    }
    /**
     * DO NOT CALL THIS METHOD, its meant as a support method for getDocs$
     */
    getPathsFromDbItemDeepRecursiveHelper(dbItem, subCollectionQueries = []) {
        if (subCollectionQueries == null || subCollectionQueries.length === 0) {
            return [dbItem.path];
        }
        let pathList = [];
        pathList.push(dbItem.path);
        subCollectionQueries.forEach(col => {
            if (Array.isArray(dbItem[col.name])) { /* property is array so will contain multiple docs */
                const docs = dbItem[col.name];
                docs.forEach(d => {
                    if (col.subCollectionQueries) {
                        pathList = pathList.concat(this.getPathsFromDbItemDeepRecursiveHelper(d, col.subCollectionQueries));
                    }
                    else {
                        pathList.push(d.path);
                    }
                });
            }
            else { /* not an array so a single doc*/
                if (col.subCollectionQueries) {
                    pathList = pathList.concat(this.getPathsFromDbItemDeepRecursiveHelper(dbItem, col.subCollectionQueries));
                }
                else {
                    const item = dbItem[col.name];
                    if (item != null && 'path' in item) {
                        pathList.push(item.path);
                    }
                    // const path = (dbItem[col.name] as FirestoreItem).path;
                }
            }
        });
        // console.log(pathList);
        return pathList;
    }
    /**
     * DO  NOT  CALL THIS METHOD, used in addDeep and updateDeep to split the data into currentDoc and subCollections
     * Onyl goes one sub level deep;
     */
    splitDataIntoCurrentDocAndSubCollections(data, subCollectionWriters = []) {
        /* Split data into current doc and sub collections */
        let currentDoc = {};
        const subCollections = {};
        /* Check if the key is in subcollections, if not place it in currentDoc, else subCollections*/
        for (const [key, value] of Object.entries(data)) {
            // console.log(key, value);
            if (subCollectionWriters) {
                const subCollectionWriter = subCollectionWriters.find(subColl => subColl.name === key);
                if (subCollectionWriter) {
                    subCollections[key] = value;
                }
                else {
                    currentDoc[key] = value;
                }
            }
            else {
                currentDoc = data;
            }
        }
        return {
            currentDoc,
            subCollections
        };
    }
    /**
     * Turn a batch into an Observable instead of Promise.
     *
     * For some reason angularfire returns a promise on batch.commit() instead of an observable like for
     * everything else.
     *
     * This method turns it into an observable
     */
    batchCommit$(batch) {
        return of(null).pipe(mergeMap(() => {
            let res$;
            res$ = from(batch.commit());
            return this.handleRes$(res$, 'batchCommit', { batch }, 'Updated', 'Failed to Update Documents').pipe(
            // take(1) /* no need to unsub*/
            );
        }), take(1));
    }
    /**
     *
     * TODO make this user defined via callback
     *
     */
    handleRes$(result$, name, data, successMessage, errorMessage) {
        return result$.pipe(
        // tap(val => console.log(val)),
        tap(() => {
            // if (successMessage != null) { (this.snackBar.open('Success', successMessage, {duration: 1000})); }
        }), 
        // map(() => true),
        catchError((err) => {
            console.error(name, data, err);
            // if (errorMessage != null) { this.snackBar.open('Error', errorMessage + '-' +  err.code, {duration: 1000}); }
            throw err;
        }));
    }
}

/** AngularfirestoreWrapperModule */
class AngularFirestoreDeepModule {
}
AngularFirestoreDeepModule.decorators = [
    { type: NgModule, args: [{
                declarations: [],
                imports: [
                    DragDropModule,
                    AngularFirestoreModule,
                ],
                exports: []
            },] }
];

/*
 * Public API Surface of angularfirestore-deep.ts
 */

/**
 * Generated bundle index. Do not edit.
 */

export { AngularFirestoreDeep, AngularFirestoreDeepModule, DocNotExistAction, addCreatedBy, addCreatedDate, addModifiedDate, convertTimestampToDate, getCollectionFsFromPath, getDocFsFromPath, getDocFsFromRef, getParentDocFsFromPath, getRefFromPath, isDocExist$ };
//# sourceMappingURL=angularfirestore-deep.js.map
